"""
pychurchtools
=============

This is the entrypoint for the ChurchTools API.

Currently, only `GET`-type methods are implemented (-> read-only).
In the future, maybe `POST`, `DELETE`, etc. may also be of interest.

TODO (for 100% completion):

    - Absence
    - Admin
    - Calendar
    - Campus
    - Checkin
    - Chat
    - Contact Label
    - Field
    - File
    - Finance
    - Groups
    - GroupHomepage
    - Job
    - Masterdata
    - Permission
    - Resource
    - Sync
    - Tag
    - Translation
    - Queue
    - Widgets

"""

import json
import traceback
from typing import Any, Dict, Optional
from urllib.parse import urljoin

import requests

from churchtools.calendars import Calendars
from churchtools.departments import Departments
from churchtools.events import Events
from churchtools.general import General
from churchtools.persons import Persons
from churchtools.services import Services
from churchtools.songs import Songs
from churchtools.status import Status
from churchtools.wiki import Wiki


class ChurchTools:

    __base_url = ""
    __debugging = 0
    __cookie = None

    def __init__(self, base_url: str, cookie: Dict[str, str] = None) -> None:
        """Initialize a CT object.

        The login cookie can look like this:

        .. code-block ::

            cookie = {
                'ChurchTools_ct_<church_name>': 'some_random_text',
            }

        :param base_url: Set the base_url to the API. It typically looks like
            this: https://<church_name>.church.tools/api/
        :type base_url: str
        :param cookie: The login cookie
        :type cookie: dict
        :returns: An initialized CT object
        :rtype: CT
        """

        self.__base_url = base_url
        if cookie:
            self.__cookie = cookie
        else:
            self.__cookie = None

        self.calendars = Calendars(self)
        self.departments = Departments(self)
        self.events = Events(self)
        self.general = General(self)
        self.persons = Persons(self)
        self.services = Services(self)
        self.songs = Songs(self)
        self.status = Status(self)
        self.wiki = Wiki(self)

    def make_request(
        self,
        endpoint: str,
        params: Any = None,
        binary: bool = False,
        method: str = "get",
    ) -> Any:
        """Make a request to the churchtools API.

        :param endpoint: The url endpoint (excluding base_url), that is called.
        :type endpoint: str
        :param params: Parameters that are to be sent along
        :type params: list, dict, or str
        :param binary: Return the binary answer.
        :type binary: bool
        :param method: The http request method
        :type method: str
        :returns: The result of the request
        :rtype: binary, dict, or string
        """

        # TODO: implement lists in params (e.g. ids for songs.list())

        rurl = urljoin(self.__base_url, "api/") + endpoint

        if self.__debugging > 1:
            print("request to:", rurl)
            for line in traceback.format_stack():
                print(line.strip())

        param_str = "?"
        if isinstance(params, dict):
            for key, value in params.items():
                if isinstance(value, list):
                    for item in value:
                        param_str += f"{key}[]={item}&"
                else:
                    param_str += f"{key}={value}&"

        if param_str[-1] == "&":
            param_str = param_str[:-1]

        if binary:
            return requests.get(rurl, params=params, cookies=self.__cookie).content

        if method == "post":
            resp = requests.post(rurl, params=params, cookies=self.__cookie)
        else:
            if param_str:
                resp = requests.get(rurl + param_str, cookies=self.__cookie)
            else:
                resp = requests.get(rurl, params=params, cookies=self.__cookie)

        rstr = resp.content.decode()

        r_ok = self.check_response(resp)
        if not r_ok:
            if self.__debugging > 0:
                print(rurl, "->", resp.status_code)
                print(rstr)

            return None

        robj = json.loads(rstr)

        if self.__debugging > 0:
            print(rurl, "->", resp.status_code)
            print(rstr)
            print()

        return robj

    def check_response(self, resp: requests.models.Response) -> bool:
        """Check a response for errors.

        :param resp: The API response
        :type resp: requsts.models.Response
        :returns: Status of the response (successful/not successful)
        :rtype: bool
        """

        if resp.status_code == 400:
            if self.__debugging > 0:
                print("Error: 400 No Authorization")
            return False
        if resp.status_code == 401:
            if self.__debugging > 0:
                print("Error: 401 Unauthorized")
            return False
        if resp.status_code == 403:
            if self.__debugging > 0:
                print("Error: 403 Forbidden")
            return False
        if resp.status_code == 404:
            if self.__debugging > 0:
                print("Error: 404 Not Found")
            return False

        return True

    def login(self, username: str, password: str, remember_me: bool = False) -> bool:
        """Login to the ChurchTools API.

        The login cookie is saved to the CT object.

        BE CAREFUL what you do with your passwords!

        :param username: The user name for the ChurchTools login
        :type username: str
        :param password: Password for the ChurchTools login
        :type password: str
        :returns: Success
        :rtype: bool
        """

        rurl = urljoin(self.__base_url, "api/login")

        params: Dict[str, Any] = {}
        params["password"] = password
        params["username"] = username
        params["rememberMe"] = remember_me
        response = requests.post(rurl, data=params, cookies=self.__cookie)

        if not self.check_response(response):
            if self.__debugging > 0:
                print("login unsuccessful")
            return False

        if self.__debugging > 0:
            print("login successful")

        cookie = response.cookies.items()[0]
        self.__cookie = {cookie[0]: cookie[1]}

        return True

    def is_authenticated(self) -> bool:
        """Check if the current object is authenticated with curchtools.

        :returns: If the object is authenticated
        :rtype: bool
        """

        # check if no cookie is set
        if not self.__cookie:
            return False

        # check if the user is 'anonymous'
        i_am = self.general.whoami()
        if i_am.id < 0:
            return False

        return True

    def set_base_url(self, base_url: str) -> None:
        """Set the base URL.

        The URL has to point to a churchtools instance.

        :param base_url: The new base_url
        :type base_url: str
        """

        self.__base_url = base_url

    def set_debugging_level(self, level: int) -> None:
        """Set the debug level.

        0: No debugging messages, only error message
        1: Some debugging messages, including full ChurchTools response
        2: Very many debugging messages, including stacktrace for every
           API query

        :param level: The debug level (0 - 2)
        :type level: int
        """

        self.__debugging = level

    def set_login_cookie(self, cookie: dict) -> None:
        """Set the login cookie.

        This cookie will be sent with every API query.
        Alternatively, you can login using func:`CT.login()`, where the cookie
        is set automatically

        Example:

        .. code-block ::

            cookie = {
                'ChurchTools_ct_<church_name>': 'some_random_text',
            }

        :param cookie: The login cookie
        :type cookie: dict
        """

        self.__cookie = cookie

    def get_login_cookie(self) -> Optional[dict]:
        """Get the login cookie.

        :returns: the login cookie
        :rtype: dict
        """

        return self.__cookie



from datetime import date, datetime
from typing import Generator, Generic, List, Optional, TypeVar, Union

from pydantic import BaseModel  # type: ignore
from pydantic.fields import ModelField

PydanticField = TypeVar("PydanticField")


class EmptyStrToNone(Generic[PydanticField]):
    @classmethod
    def __get_validators__(cls) -> Generator:
        yield cls.validate

    @classmethod
    def validate(cls, v: PydanticField, field: ModelField) -> Optional[PydanticField]:
        if v == "":
            return None
        return v


class EmptyStrToFalse(Generic[PydanticField]):
    @classmethod
    def __get_validators__(cls) -> Generator:
        yield cls.validate

    @classmethod
    def validate(cls, v: PydanticField, field: ModelField) -> bool:
        if v == "" or not v:
            return False
        return True


class Person(BaseModel):

    id: int
    securityLevelForPerson: int
    editSecurityLevelForPerson: int
    title: Optional[str]
    firstName: Optional[str]
    lastName: Optional[str]
    nickname: Optional[str]
    job: Optional[str]
    street: Optional[str]
    addressAddition: Optional[str]
    zip: Optional[str]
    city: Optional[str]
    country: Optional[str]
    latitude: Optional[float]
    longitude: Optional[float]
    latidudeLoose: Optional[float]
    longitudeLoose: Optional[float]
    phonePrivate: Optional[str]
    phoneWork: Optional[str]
    mobile: Optional[str]
    fax: Optional[str]
    birthName: Optional[str]
    birthday: Optional[EmptyStrToNone[date]]
    birthplace: Optional[str]
    imageUrl: Optional[str]
    familyImageUrl: Optional[str]
    sexId: Optional[int]
    email: Optional[str]
    cmsUserId: Optional[str]
    optigemUserId: Optional[str]
    nationalityId: Optional[int]
    familyStatusId: Optional[int]
    weddingDate: Optional[EmptyStrToNone[date]]
    campusId: Optional[int]
    statusId: Optional[int]
    departmentIds: Optional[List[int]]
    firstContact: Optional[EmptyStrToNone[datetime]]
    dateOfBelonging: Optional[EmptyStrToNone[date]]
    dateOfEntry: Optional[EmptyStrToNone[datetime]]
    dateOfResign: Optional[EmptyStrToNone[datetime]]
    dateOfBaptism: Optional[EmptyStrToNone[date]]
    placeOfBaptism: Optional[str]
    baptisedBy: Optional[str]
    referredBy: Optional[str]
    referredTo: Optional[str]
    growPathId: Optional[int]
    isArchived: Optional[bool]

    # TODO: privacyPolicyAgreement
    # TODO: meta

    def __repr__(self) -> str:

        return f"<Person: {self.firstName} {self.lastName} [{self.id}]>"


class AgendaSong(BaseModel):

    songId: int
    arrangementId: int
    title: str
    arrangement: str
    category: str
    key: str
    bpm: str
    defaultArrangement: Optional[str]


class AgendaItem(BaseModel):

    id: int
    position: int
    type: str
    title: str
    note: Optional[str]
    duration: int
    start: datetime
    isBeforeEvent: bool
    song: Optional[AgendaSong]

    # TODO: responsible
    # TODO: serviceGroupNotes
    # TODO: meta


class Agenda(BaseModel):

    id: int
    name: str
    series: str
    isFinal: bool
    calendarId: int
    total: int
    items: List[AgendaItem]

    # TODO: meta

    def __repr__(self) -> str:

        return f"<Agenda: {self.name} [{self.id}]>"


class Service(BaseModel):

    id: int
    name: str
    serviceGroupId: int
    commentOnConfirmation: bool
    sortKey: int
    allowDecline: bool
    allowExchange: bool
    comment: str
    standard: bool
    hidePersonName: bool
    sendReminderMails: bool
    sendServiceRequestMails: Optional[bool]
    allowControlLiveAgenda: bool
    groupIds: Optional[str]
    tagIds: Optional[str]
    calTextTemplate: str
    allowChat: bool

    def __repr__(self) -> str:

        return f"<Service: {self.name} [{self.id}]>"


class ServiceGroup(BaseModel):

    id: int
    name: str
    sortKey: int
    viewAll: Optional[bool]
    campusId: Optional[int]
    onlyVisibleInCampusFilter: Optional[bool]

    def __repr__(self) -> str:

        return f"<ServiceGroup: {self.name} [{self.id}]>"


class EventService(BaseModel):

    id: int
    name: Optional[str]
    serviceId: int
    agreed: bool
    isValid: bool
    requestedDate: datetime
    comment: str
    counter: int
    allowChat: bool

    # TODO: person
    # TODO: requesterPerson

    def __repr__(self) -> str:

        return f"<EventService: {self.name}>"


class Event(BaseModel):

    id: int
    guid: str
    name: str
    description: str
    startDate: datetime
    endDate: datetime
    chatStatus: str
    eventServices: Optional[List[EventService]]

    # TODO: permissions
    # TODO: calendar

    def __repr__(self) -> str:

        return (
            f"<Event: {self.startDate.day:02}."
            f"{self.startDate.month:02} {self.startDate.hour:02}:"
            f'{self.startDate.minute:02} "{self.name}" [{self.id}]>'
        )


class GroupDomainAttributes(BaseModel):

    note: str


class Group(BaseModel):

    title: str
    domainType: str
    domainIdentifier: str
    apiUrl: str
    frontendUrl: str
    imageUrl: Optional[str]
    domainAttributes: GroupDomainAttributes

    def __repr__(self) -> str:

        return f"<Group: {self.title}>"


class ArrangementLink(BaseModel):

    domainType: str
    domainId: str
    name: str
    filename: str
    fileUrl: str


class ArrangementFile(BaseModel):

    domainType: str
    domainId: str
    name: str
    filename: str
    fileUrl: str


class Arrangement(BaseModel):

    id: int
    name: str
    isDefault: bool
    keyOfArrangement: str
    bpm: str
    beat: str
    duration: str
    note: str
    links: List[ArrangementLink]
    files: List[ArrangementFile]

    # TODO: meta


class SongCategory(BaseModel):

    id: int
    name: str
    nameTranslated: str
    sortKey: int
    campusId: Optional[int]


class Song(BaseModel):

    id: int
    name: str
    category: SongCategory
    shouldPractice: bool
    author: str
    ccli: str
    copyright: str
    note: str
    arrangements: List[Arrangement]

    # TODO: meta

    def __repr__(self) -> str:

        return f"<Song: {self.name} [{self.id}]>"


class VersionInfo(BaseModel):

    build: str
    version: str

    def __repr__(self) -> str:

        return f"<Version: {self.version} (build: {self.build})>"


class PersonTag(BaseModel):

    id: int
    name: str
    count: int


class PersonDomainAttributes(BaseModel):

    firstName: str
    lastName: str
    guid: str


class PersonDomainObject(BaseModel):

    domainAttributes: PersonDomainAttributes
    imageUrl: str
    frontendUrl: str
    apiUrl: str
    domainType: str
    title: str
    domainIdentifier: str


class PersonRelationship(BaseModel):

    relative: PersonDomainObject
    degreeOfRelationship: str
    relationshipName: str
    relationshipTypeId: int

    def __repr__(self) -> str:

        return (
            f"<Relationship: {self.relative.title} " f"({self.degreeOfRelationship})>"
        )


class Setting(BaseModel):

    module: str
    attribute: str
    value: Union[dict, int, List, str]

    def __repr__(self) -> str:

        return f"<Setting: {self.module}.{self.attribute} = {self.value}>"


class CTStatus(BaseModel):

    id: int
    name: str
    shorty: str
    isMember: bool
    isSearchable: bool
    sortKey: int
    securityLevelId: int


class WikiCategory(BaseModel):

    id: int
    name: str
    sortKey: int
    campusId: Optional[int]
    inMenu: bool
    fileAccessWithoutPermission: bool
    nameTranslated: str

    def __repr__(self) -> str:

        return f"<WikiCategory: {self.name} [{self.id}]>"


class WikiPermission(BaseModel):

    canEdit: bool


class WikiPage(BaseModel):

    identifier: str
    wikiCategory: WikiCategory
    title: str
    version: int
    text: Optional[str]
    onStartpage: EmptyStrToFalse[bool]
    redirectTo: Optional[str]
    permissions: WikiPermission
    isMarkdown: bool

    # TODO: meta

    def __repr__(self) -> str:

        return f"<WikiPage: {self.title} [{self.identifier}]>"


class WikiSearchResult(BaseModel):

    title: str
    domainType: str
    domainIdentifier: str
    apiUrl: str
    frontendUrl: str
    imageUrl: Optional[str]
    preview: str

    def __repr__(self) -> str:

        return f"<WikiSearchResult: {self.title} ({self.preview})>"


class BirthdayPersonDomainAttributes(BaseModel):

    firstName: str
    lastName: str
    guid: str


class BirthdayPerson(BaseModel):

    title: str
    domainType: str
    domainIdentifier: str
    apiUrl: str
    frontendUrl: str
    imageUrl: Optional[str]
    domainAttributes: BirthdayPersonDomainAttributes

    def __repr__(self) -> str:

        return f"<BirthdayPerson: {self.title}>"


class Birthday(BaseModel):

    type: str
    date: date
    age: Optional[int]
    person: BirthdayPerson

    def __repr__(self) -> str:

        return f"<Birthday: {self.person.title} {self.date} " f"({self.age} years)>"


class ServiceRequest(BaseModel):

    id: int
    personId: int
    name: str
    serviceId: int
    agreed: bool
    isValid: bool
    requestedDate: datetime
    requesterPersonId: int
    comment: str
    counter: int

    # TODO: person
    # TODO: requesterPerson

    def __repr__(self) -> str:

        return f"<ServiceRequest: {self.name} {self.serviceId} [{self.id}]>"


class Device(BaseModel):

    id: str
    type: str
    ttl: datetime
    version: str
    createdAt: datetime
    updatedAt: datetime


class Pagination(BaseModel):

    total: int
    current: int
    limit: int
    lastPage: int


class MetaPagination(BaseModel):

    count: int
    pagination: Pagination


class Department(BaseModel):

    id: int
    name: str
    nameTranslated: Optional[str]
    sortKey: int

    def __repr__(self) -> str:

        return f"<Department: {self.name} [{self.id}]>"


class SearchResult(BaseModel):

    apiUrl: str
    domainIdentifier: str
    domainType: str
    frontendUrl: str
    icon: str
    imageUrl: Optional[str]
    title: str

    # TODO: domainAttributes

    def __repr__(self) -> str:

        return f"<SearchResult: {self.title} [{self.domainType}]>"


class Calendar(BaseModel):

    campusId: Optional[int]
    color: str
    eventTemplateId: Optional[int]
    iCalSourceUrl: Optional[str]
    id: int
    isPrivate: bool
    isPublic: bool
    name: str
    nameTranslated: str
    randomUrl: str
    sortKey: int

    # TODO: meta


class Address(BaseModel):

    meetingAt: Optional[str]
    street: Optional[str]
    addition: Optional[str]
    district: Optional[str]
    zip: Optional[str]
    city: Optional[str]
    country: Optional[str]
    latitude: Optional[float]
    longitued: Optional[float]


class Appointment(BaseModel):
    id: Union[int, str]
    caption: str
    note: str
    address: Optional[Address]
    version: int
    calendar: Calendar
    information: Optional[str]
    image: Optional[str]
    link: str
    isInternal: bool
    startDate: Union[date, datetime]
    endDate: Union[date, datetime]
    allDay: bool
    repeatId: int
    repeatFrequency: Optional[int]
    repeatUntil: Optional[str]
    repeatOption: Optional[int]

    # TODO: additions, exceptions, meta



"""
General
=======

Endpoints of general purpose.

"""

from churchtools.ct_types import Person, SearchResult, VersionInfo
from typing import Any, Dict, List


class General:
    def __init__(self, ct: Any) -> None:
        """Initialize a General object."""

        self.__ct = ct

    def csrftoken(self) -> str:
        """Returns the CSRF-Token for the current user.

        :returns: CSRF-Token for the current user
        :rtype: str
        """

        res = self.__ct.make_request("csrftoken")
        return res["data"]

    def info(self) -> VersionInfo:
        """Information about the API.

        The API envoles and dependes on the ChurchTools version.
        This endpoint provides the build version and CT version.

        :returns: The version info and build number
        :rtype: VersionInfo
        """

        res = self.__ct.make_request("info")
        return VersionInfo(**res)

    def whoami(self) -> Person:
        """Currently logged in user.

        This endpoint returns the current user.
        If the request is unauthorized, the anonymous user (aka public user)
        is returned.

        .. note:: In the API there is a parameter defined
            (only_allow_authenticated). We ignore this, as it has no use here.

        :returns: The current user, that is using the API
        :rtype: Person
        """

        res = self.__ct.make_request("whoami")
        return Person(**res["data"])

    def search(
        self,
        query: str,
        domain_types: List[str] = ["person", "group", "song", "wiki_page"],
    ) -> List[SearchResult]:
        """Search globally for different or all domain types.

        :param query: The search query
        :type query: str
        :param domain_types: A list of domain types to query. Possible values
          are 'person', 'group', 'song', 'wiki_page'
        :returns: A list of search results
        :rtype: SearchResult
        """

        params: Dict[str, Any] = {}
        params["query"] = query
        params["domainTypes[]"] = domain_types

        res = self.__ct.make_request("search", params=params)

        results = []
        if res and "data" in res:
            for item in res["data"]:
                result = SearchResult(**item)
                results.append(result)

        return results

    def test_route(self, route: str) -> dict:
        """Test an arbitrary route."""

        res = self.__ct.make_request(route)
        return res
